__version__ = "1.0.0.1b3"

from .fast_gp_lattice import FastGPLattice
from .fast_gp_digital_net_b2 import FastGPDigitalNetB2
from .standard_gp import StandardGP
from .plots import plot_fastgps_fit_data
